-- Puffin: A Tojaso FowlPlay Project
-- Component containing class functions shared by view components

-- Views are defined by declarations in the shared database, including position information,
-- widget templates and layouts, plus data providers for updates.
-- Views are responsible for registering for callbacks in order to keep themselves updated.

local _, Puffin = ...
local EF = Puffin.Error
local DB = Puffin.Database

local class = Puffin.Class

local View = class("View")
Puffin.Classes.View = View

local classificationColors = { rare = "silver", elite = "gold", rareelite = "gold", worldboss = "orange" }
local classificationString = { rare = "|cFFC0C0C0+", elite = "|CFFFFD700+", rareelite = "|CFFFFD700+", worldboss = "|CFFFFA500+" }

-- Initialize a view based on information in the description table
function View:initialize(name, dt)
	local layout = DB.GetRecord("layouts", dt.layout) -- get layout from table of widget layouts
	assert(layout, EF("View", "widget layout not found"))
	assert(DB.IsDatabase("templates"), EF("View", "widget templates not found"))

	self.name = name -- save the name of the view
	self.widgets = {} -- table of widgets used by the view
	for i, w in ipairs(layout) do -- iterate over list of widgets for this view
		local c = Puffin.Classes[w.class] -- get class of each widget
		assert(c, EF("View", "widget class not recognized"))
		local template = DB.GetRecord("templates", w.class)
		assert(template, EF("View", "template not found"))
		local widget = c:new(template, w.attributes) -- use template and layout to instantiate the widget
		assert(widget, EF("View", "widget not instantiated"))
		while c.super do
			c = c.super
			template = DB.GetRecord("templates", c.name)
			if template then widget:Configure(template, w.attributes) end -- apply template for each superclass
		end
		if i == 1 then
			local f = UIParent
			if dt.relativeFrame then f = _G[dt.relativeFrame] end
			widget:SetPosition(dt.point or "CENTER", f, dt.relative or "CENTER", dt.offsetX, dt.offsetY) -- position base widget relative to either a frame or UIParent
		else
			widget:SetPosition(w.point, self.widgets[1], w.relative, w.offsetX, w.offsetY) -- position relative to base widget
		end
		if dt.strata then widget:SetStrata(dt.strata) end
		self.widgets[i] = widget
		widget:Update() -- do the initial update on each widget
	end
end

-- Reconfigure a view based on information in the description table
function View:Reconfigure(dt)
	local layout = DB.GetRecord("layouts", dt.layout) -- get layout from table of widget layouts
	assert(layout, EF("View", "widget layout not found"))
	assert(DB.IsDatabase("templates"), EF("View", "widget templates not found"))

	if self.widgets then
		for i, widget in ipairs(self.widgets) do
			local c = widget.class
			local template = DB.GetRecord("templates", c.name)
			assert(template, EF("View", "template not found"))
			local w = layout[i] -- uses same index as the widgets table
			assert(w, EF("View", "layout not found"))
			widget:Configure(template, w.attributes) -- apply updated template and layout values
			while c.super do
				c = c.super
				template = DB.GetRecord("templates", c.name)
				if template then widget:Configure(template, w.attributes) end -- apply template for each superclass
			end
			if i == 1 then
				widget:SetPosition(dt.point, nil, dt.relative, dt.offsetX, dt.offsetY) -- position base widget
			else
				widget:SetPosition(w.point, self.widgets[1], w.relative, w.offsetX, w.offsetY) -- position relative to base widget
			end
			widget:Update() -- display with any changes
		end
	end
end

-- Set the frame strata for the widgets in a view
function View:SetStrata(strata)
	if self.widgets then
		for i, widget in ipairs(self.widgets) do widget:SetStrata(strata) end
	end
end

-- Set the frame level for the widgets in a view
function View:SetLevel(level)
	if self.widgets then
		for i, widget in ipairs(self.widgets) do widget:SetLevel(level) end
	end
end

-- View_Castbar: subclass of view that supports castbars

local View_Castbar = View:subclass("View_Castbar")
Puffin.Classes.View_Castbar = View_Castbar

-- Initialize view for a castbar and register for update events
function View_Castbar:initialize(name, dt)
	View.initialize(self, name, dt) -- pass name and description table to superclass initialize method
	self.unit = dt.unit -- unit associated with this castbar
	Puffin.RegisterCallback("Puffin_Casting" .. dt.unit, View_Castbar.Update, self)
	Puffin.Callback("Puffin_Casting_Unit", dt.unit, true) -- register interest in updates for associated unit's castbars
	self:Update() -- complete initialization
end

-- Update castbar for the associated unit
function View_Castbar:Update()
	local enabled = DB.GetAttribute("views", self.name, "enabled")
	local configurationMode = DB.GetAttribute("defaults", "configuration", "mode")
	local hidden = enabled and not configurationMode
	local cb = DB.GetRecord("casting", self.unit)
	if enabled and cb and cb.label then
		local barWidget = self.widgets[1]
		local iconWidget = self.widgets[2]
		local labelWidget = self.widgets[3]
		local timeWidget = self.widgets[4]
		local durationWidget = self.widgets[5]
		if barWidget then
			barWidget:Set("fgColor", cb.color)
			barWidget:Set("bgColor", cb.background)
			barWidget:SetTime(cb.expire, cb.duration, cb.timeLeft)
			barWidget:Set("fillBars", not cb.channel)
		end
		if iconWidget then
			iconWidget:SetTexture(cb.icon)
		end
		if labelWidget then
			labelWidget:SetText(cb.label)
		end
		if timeWidget then
			timeWidget:SetTime(cb.expire, cb.duration, cb.timeLeft)
		end
		if durationWidget then
			durationWidget:SetText(string.format("%.1f", cb.duration))
		end
		hidden = false
		local level = self.level or 0
		if level ~= self._level then self:SetLevel(level); self._level = level end
		local strata = self.strata or "MEDIUM"
		if strata ~= self._strata then self:SetStrata(strata); self._strata = strata end
	end
	if configurationMode then hidden = false end
	for i, widget in ipairs(self.widgets) do widget:SetHidden(hidden); widget:Update() end -- show/hide all the widgets
end

-- View_Unit: subclass of view that supports unit frames

local View_Unit = View:subclass("View_Unit")
Puffin.Classes.View_Unit = View_Unit

-- Initialize view for a unit frame and register for update events
function View_Unit:initialize(name, dt)
	self.unit = dt.unit -- unit associated with this unit frame
	-- find unit frame and use its position for this view
	View.initialize(self, name, dt) -- pass name and description table to superclass initialize method
	self:Update() -- complete initialization
	Puffin.RegisterCallback("Puffin_Update", View_Unit.Update, self)
end

-- Update unit frame widgets
function View_Unit:Update()
	local enabled = DB.GetAttribute("views", self.name, "enabled")
	local configurationMode = DB.GetAttribute("defaults", "configuration", "mode")
	local hidden = enabled and not configurationMode
	if enabled then
		local unit = self.unit
		if UnitExists(unit) then
			if unit == "player" or unit == "pet" then
				if UnitHasVehicleUI("player") then -- switch units around if player in vehicle
					if unit == "player" then unit = "pet" else unit = "player" end
				end
			end
			local classText, classToken = UnitClassBase(unit)
			local classColor = classToken and RAID_CLASS_COLORS[classToken] or nil
			local isPlayer = UnitIsPlayer(unit)
			local isPVP = UnitIsPVP(unit)
			local h = UnitHealth(unit)
			local hm = UnitHealthMax(unit)
			local p = UnitPower(unit)
			local pm = UnitPowerMax(unit)
			local raidIndex = GetRaidTargetIndex(unit)
			local healthWidget = self.widgets[1]
			local textWidget = self.widgets[3]
			local iconWidget = self.widgets[4]
			local powerWidget = self.widgets[5]
			local pvpWidget = self.widgets[6]
			local stateWidget = self.widgets[7]
			if healthWidget then
				if h == 0 then hm = 1 end -- make sure not dividing by zero
				healthWidget:SetValue(h / hm)
				local c = DB.GetAttribute("defaults", "colors", "green")
				local r, g, b = c.r, c.g, c.b
				if not isPlayer then r, g, b = UnitSelectionColor(unit) end
				c = healthWidget:Get("fgColor")
				if not c or c.r ~= r or c.g ~= g or c.b ~= b then healthWidget:Set("fgColor", { r = r, g = g, b = b, a = 1 }) end
			end
			if powerWidget then
				if p == 0 then pm = 1 end -- make sure not dividing by zero
				powerWidget:SetValue(p / pm)
				local powerType, powerToken, r, g, b = UnitPowerType(unit)
				local c = PowerBarColor[powerToken]
				if c then
					r, g, b = c.r, c.g, c.b
				elseif not r then
						c = PowerBarColor[powerType] or PowerBarColor["MANA"]
						r, g, b = c.r, c.g, c.b
				end
				c = powerWidget:Get("fgColor")
				if not c or c.r ~= r or c.g ~= g or c.b ~= b then powerWidget:Set("fgColor", { r = r, g = g, b = b, a = 1 }) end
			end
			if pvpWidget then
				pvpWidget:SetCondition(isPVP)
			end
			if iconWidget then
				if raidIndex then
					local tex = iconWidget:GetIconTexture()
					if tex then SetRaidTargetIconTexture(tex, raidIndex) end
				end
				iconWidget:SetCondition(raidIndex ~= nil)
			end
			if textWidget then
				local raceText = ""
				if isPlayer then
					raceText = UnitRace(unit) or UnitCreatureType(unit) or "Humanoid" or ""
				else
					raceText = UnitCreatureType(unit)
					if not raceText or raceText == "Not specified" then raceText = UNKNOWN end
				end
				local t = GetUnitName(unit) or "Name"
				local levelText = UnitLevel(unit) or "1"
				local classy = UnitClassification(unit)
				if levelText == -1 then
					levelText = "|cFFFF0000??|r"
				elseif classy then
					local c = classificationString[classy]
					if c then levelText = c .. levelText .. "|r" end
				end
				if unit == "player" or unit == "target" or unit == "pet" then
					t = t .. " | " .. tostring(levelText) .. " " .. raceText
					if isPlayer then t = t .. " |c" .. (classColor and classColor.colorStr or "FFFF0000") .. classText .. "|r" end
					if isPVP and unit == "player" then
						timer = GetPVPTimer()
						if timer < 300000 then
							seconds = math.floor((timer / 1000) + 0.5)
							minutes = math.floor(seconds / 60)
							seconds = seconds - (minutes * 60)
							local s = string.format("%d:%02d", minutes, seconds)
							t = t .. " | |cFFFF0000PvP " .. s .. "|r"
						end
					end
				end
				textWidget:SetText(t)
			end
			if classWidget then
				if isPlayer then
					classWidget:Set("backdropColor", classColor)
				else
					local r, g, b = UnitSelectionColor(unit)
					local c = classWidget:Get("backdropColor")
					if not c or c.r ~= r or c.g ~= g or c.b ~= b then classWidget:Set("backdropColor", { r = r, g = g, b = b, a = 1 }) end
				end
			end
			if stateWidget then
				if unit == "player" then -- for player this square is status
					local c = false
					if IsResting() then c = DB.GetAttribute("defaults", "colors", "lightBlue") end
					if UnitAffectingCombat(unit) then c = DB.GetAttribute("defaults", "colors", "transparentRed") end
					if UnitIsDeadOrGhost(unit) then c = DB.GetAttribute("defaults", "colors", "transparentWhite") end
					if not UnitIsConnected(unit) then c = DB.GetAttribute("defaults", "colors", "transparentBlack") end
					if c then stateWidget:Set("backdropColor", c) end
					stateWidget:SetCondition(c ~= false)
				else
					stateWidget:SetCondition(false)
				end
			end
			hidden = false
			local level = self.level or 0
			if level ~= self._level then self:SetLevel(level); self._level = level end
			local strata = self.strata or "MEDIUM"
			if strata ~= self._strata then self:SetStrata(strata); self._strata = strata end
		else hidden = true end
	end
	if configurationMode then hidden = false end
	for i, widget in ipairs(self.widgets) do widget:SetHidden(hidden); widget:Update() end -- show/hide and update all the widgets
end

-- View_Nameplate: subclass of view that supports nameplates

local View_Nameplate = View:subclass("View_Nameplate")
Puffin.Classes.View_Nameplate = View_Nameplate

-- Initialize view for a unit frame and register for update events
function View_Nameplate:initialize(name, dt)
	self.unit = false
	self.strata = dt.strata
	self.nameplate = dt.nameplate
	-- Puffin.Debug("View_Nameplate", name, dt.layout, dt.relativeFrame)
	View.initialize(self, name, dt) -- pass name and description table to superclass initialize method
	self:Update() -- complete initialization
	Puffin.RegisterCallback("Puffin_Update", View_Nameplate.Update, self)
	Puffin.RegisterCallback("Puffin_Refresh", View_Nameplate.Refresh, self)
end

function View_Nameplate:SetPosition(w)
	local World, Align = Puffin.World, Puffin.Align
	local x, y = self.nameplate:GetCenter() -- get position of underlying nameplate
	local tc = w._display.container
	local dx = World(x) - (tc:GetWidth() / 2); local dy = World(y) - tc:GetHeight()
	tc:ClearAllPoints(); tc:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", Align(dx), Align(dy))
end

-- Update nameplate widgets
function View_Nameplate:Update()
	local hidden = true
	local unit = self.unit
	if unit and UnitExists(unit) then
		local t, c, r, g, b
		local isPlayer = UnitIsPlayer(unit)
		local isTarget = UnitIsUnit("target", unit)
		local isCombat = UnitAffectingCombat(unit)
		local h = UnitHealth(unit)
		local hm = UnitHealthMax(unit)
		local raidIndex = GetRaidTargetIndex(unit)
		local w = self.widgets[1] -- custom nameplate widget
if IsAltKeyDown() then isCombat = true end
		if w then
			local z = w.bar
			if z then -- nameplate has a single bar that displays health
				if h == 0 then hm = 1 end -- make sure not dividing by zero
				z.value = h / hm
				r, g, b = 0, 1, 0 -- use green for players
				if not isPlayer then r, g, b = UnitSelectionColor(unit) end
				c = z.fgColor
				if not c or c.r ~= r or c.g ~= g or c.b ~= b then z.fgColor = { r = r, g = g, b = b, a = 1 } end
				if isCombat then z.height = 10 else z.height = 4 end
			end
			
			if w.texts then -- nameplate has a name string and a conditionally shown level string
				z = w.texts.name
				if z then z.text = GetUnitName(unit) or "Unknown" end

				z = w.texts.level
				if z then
					t = UnitLevel(unit) or "1"
					local classy = UnitClassification(unit)
					if t == -1 then
						t = "|cFFFF0000??|r"
					elseif classy then
						c = classificationString[classy]
						if c then t = c .. t .. "|r" end
					end
					if t == UnitLevel("player") then t = "" end
					z.text = t
				end
			end
			
			if w.icons then -- nameplate has a conditionally shown raid target icon
				z = w.icons.raidTarget
				if z then z.raidIndex = raidIndex; z.condition = (raidIndex ~= nil) end
			end
			
			if w.panels then -- nameplate has rectangular indicators for selection and threat
				z = w.panels.targetIndicator
				if z then
					z.condition = isTarget
				end
				
				z = w.panels.threatIndicator
				if z then
					t = UnitThreatSituation("player", unit)
					if t and t > 0 then
						r, g, b = GetThreatStatusColor(t)
						z.backdropColor = { r = r, g = g, b = b, a = 1 }
						z.condition = true
					else
						z.condition = false
					end
				end
			end
			
			if w.auras then -- nameplate shows auras to help track buffs on friends and debuffs on enemies
				z = w.auras
				local count = 0
				local filter = UnitCanAttack("player", unit) and "PLAYER|HARMFUL" or "PLAYER|HELPFUL"
				for i = 1, 40 do
					local name, _, icon, _, _, duration, expires = UnitAura(unit, i , filter)
					if not name then break end
					count = count + 1
					local pt = z.frames[count]
					pt.texture = icon
					if duration > 0 then pt.timeLeft = expires - GetTime() else pt.timeLeft = nil end
					if count >= z.maxAuras then break end
				end
				if count > 1 then -- more than one aura so must sort them
					for i = 1, count do
						local za = z.frames[i]
						for j = 2, count do
							local zb = z.frames[j]
							if (za.timeLeft or 9999) > (zb.timeLeft or 9999) then -- bubble sort, swap to get shorter times to front
								local zt, zi = za.timeLeft, za.texture
								za.timeLeft = zb.timeLeft; za.texture = zb.texture; zb.timeLeft = zt; zb.texture = zi
							end
						end
					end
				end
				z.activeAuras = count
			end
		end

		hidden = false
		local level = isTarget and 500 or self.level
		if level ~= self._level then self:SetLevel(level); self._level = level end -- target always is highest level
		local strata = self.strata or "BACKGROUND"
		if strata ~= self._strata then self:SetStrata(strata); self._strata = strata end
		self:SetPosition(w) -- set position relative to the hidden blizzard nameplate
	end
	for i, widget in ipairs(self.widgets) do widget:SetHidden(hidden); widget:Update() end -- show/hide and update all the widgets
end

-- Refresh nameplate position
function View_Nameplate:Refresh()
	local unit = self.unit
	if unit and UnitExists(unit) then
		local w = self.widgets[1]
		if w then self:SetPosition(w) end
	end
end

-- View_Resources: subclass of view that supports the display of player resources (e.g., combo points)

local View_Resources = View:subclass("View_Resources")
Puffin.Classes.View_Resources = View_Resources

local classResources = { -- power tokens used to look up colors in PowerBarColor global table for classes supported by the resource bar
	DRUID = "COMBO_POINTS", ROGUE = "COMBO_POINTS", WARLOCK = "SOUL_SHARDS", MAGE = "ARCANE_CHARGES", MONK = "CHI", PALADIN = "HOLY_POWER",
}

-- Initialize view for a unit frame and register for update events
function View_Resources:initialize(name, dt)
	local _, myClass = UnitClassBase("player") -- check if this class uses resources
	local cp = classResources[myClass]
	if cp then -- only enable the resource bar for classes with entries in this table
		View.initialize(self, name, dt) -- pass name and description table to superclass initialize method
		self._myClass = myClass
		self._myPower = _G["SPELL_POWER_" .. cp] -- this is the power type number passed to UnitPower
		if cp then cp = PowerBarColor[cp] else cp = RAID_CLASS_COLORS[myClass] end
		self._myColor = { r = cp.r, g = cp.g, b = cp.b, a = 1 } -- copy with slight transparency
		self._extraColor = { r = 1, g = 0, b = 0, a = 1 }
		self._emptyColor = { r = 0.25, g = 0.25, b = 0.25, a = 1 }
		-- for k, v in pairs(PowerBarColor) do Puffin.Debug("pcolor", k, tostring(v.r), tostring(v.g), tostring(v.b)) end
		self:Update() -- complete initialization
		Puffin.RegisterCallback("Puffin_Update", View_Resources.Update, self)
	end
end

-- Adjust sizes and positions of widgets for the number of available resources (currently 3, 4, 5 or 6)
local function Configure_Resources(maxcp, widgets)
	local t = widgets[1] -- contains the size/position for various numbers of resources
	if maxcp > 6 then maxcp = 5 end -- note that reverting to 5 if greater than 6 is intentional
	local inset = t.panel_inset[maxcp]
	local spacing = t.panel_spacing[maxcp]
	local width = t.panel_width[maxcp]
	local height = t.panel_height
	local w
	if maxcp == 3 then
		w = widgets[2]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset, height)
		w = widgets[3]; w:Set("width", width); w:SetPosition("CENTER", t, "CENTER", 0, height)
		w = widgets[4]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset, height)
	elseif maxcp == 4 then
		w = widgets[2]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset, height)
		w = widgets[3]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset + width + spacing, height)
		w = widgets[4]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset - width - spacing, height)
		w = widgets[5]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset, height)
	elseif maxcp == 5 then
		w = widgets[2]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset, height)
		w = widgets[3]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset + width + spacing, height)
		w = widgets[4]; w:Set("width", width); w:SetPosition("CENTER", t, "CENTER", 0, height)
		w = widgets[5]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset - width - spacing, height)
		w = widgets[6]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset, height)
	elseif maxcp == 6 then
		w = widgets[2]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset, height)
		w = widgets[3]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset + width + spacing, height)
		w = widgets[4]; w:Set("width", width); w:SetPosition("LEFT", t, "LEFT", inset + 2 * (width + spacing), height)
		w = widgets[5]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset - 2 * (width + spacing), height)
		w = widgets[6]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset - width - spacing, height)
		w = widgets[7]; w:Set("width", width); w:SetPosition("RIGHT", t, "RIGHT", -inset, height)
	end
	for i = 1, 6 do widgets[i+1]:SetCondition(i <= maxcp) end
end

-- Update resource-related widgets
function View_Resources:Update()
	local enabled = DB.GetAttribute("views", self.name, "enabled")
	local configurationMode = DB.GetAttribute("defaults", "configuration", "mode")
	local hidden = enabled and not configurationMode
	if enabled then
		local myClass = self._myClass
		local myPower = self._myPower
		local cpColor = self._myColor -- combo pts up to 5 are colored with class or power type color
		local apColor = self._extraColor -- "extra" combo pts are red (e.g., see rogue Anticipation)
		local npColor = self._emptyColor -- backdrop color for empty slot
		if (myClass == "ROGUE") or (myClass == "DRUID" and (GetShapeshiftFormID() == 1)) then -- rogue or druid in cat form use combo points
			local maxcp = UnitPowerMax("player", myPower)
			if self._maxcp ~= maxcp then Configure_Resources(maxcp, self.widgets); self._maxcp = maxcp end -- this will reposition/resize the resource widgets
			local cp = UnitPower("player", myPower)
			if maxcp == 6 then
				for i = 1, 6 do
					local w = self.widgets[i + 1]
					w:Set("backdropColor", (i <= cp) and cpColor or npColor)
				end
			else
				for i = 1, 5 do
					local w = self.widgets[i + 1]
					w:Set("backdropColor", (i <= cp) and ((i <= (cp - 5)) and apColor or cpColor) or npColor)
				end
			end
			hidden = false
		elseif (myClass == "WARLOCK") or (myClass == "MONK" and GetSpecialization() == SPEC_MONK_WINDWALKER) or  -- other classes with special resources
			(myClass == "MAGE" and GetSpecialization() == SPEC_MAGE_ARCANE) or (myClass == "PALADIN" and GetSpecialization() == SPEC_PALADIN_RETRIBUTION) then
			local maxcp = UnitPowerMax("player", myPower)
			if self._maxcp ~= maxcp then Configure_Resources(maxcp, self.widgets); self._maxcp = maxcp end -- this will reposition/resize the resource widgets
			local cp = UnitPower("player", myPower) or 0 -- all specs use soul shards now
			for i = 1, maxcp do
				local w = self.widgets[i + 1]
				w:Set("backdropColor", (i <= cp) and cpColor or npColor)
			end
			hidden = false
		else hidden = true end
	end
	if configurationMode then hidden = false end
	for i, widget in ipairs(self.widgets) do widget:SetHidden(hidden); widget:Update() end -- show/hide and update all the widgets
end

-- View_Minimap: subclass of view that supports the minimap with placement and a border

local View_Minimap = View:subclass("View_Minimap")
Puffin.Classes.View_Minimap = View_Minimap

-- Initialize view for the minimap
function View_Minimap:initialize(name, dt)
	local layout = DB.GetRecord("layouts", dt.layout) -- get layout from table of widget layouts
	assert(layout, EF("View_Minimap", "widget layout not found"))
	if dt.scale then Minimap:SetScale(dt.scale) end
	local Align = Puffin.Align -- function used to align pixel perfectly
	local Pixel = Puffin.Pixel -- function used to scale to pixel dimensions
	assert(type(Align) == "function", EF("View_Minimap", "align function missing"))
	assert(type(Pixel) == "function", EF("View_Minimap", "pixel function missing"))
	local w = Align(Minimap:GetWidth())
	local h = Align(Minimap:GetHeight())
	Minimap:SetSize(w, h)
	Minimap:ClearAllPoints()
	Minimap:SetPoint(dt.point or "CENTER", nil, dt.relative or "CENTER", Pixel(dt.offsetX or 0), Pixel(dt.offsetY or 0)) -- place the minimap frame first
	local x = layout[1] -- only one widget in the layout
	x.attributes.scale = dt.scale -- applies scale factor, if any, to widgets
	x.attributes.width = Puffin.Screen(w); x.attributes.height = Puffin.Screen(h) -- adjust width and height to match the minimap
	View.initialize(self, name, dt) -- pass name and description table to superclass initialize method
	local panel = self.widgets[1]
	assert(panel, EF("View_Minimap", "panel widget not found"))
	panel:SetPosition("CENTER", Minimap, "CENTER") -- anchor panel to the minimap frame
end

View.static.activeViews = {} -- class variable that contains all active views

-- Class method to create and configure views based on configuration info in the shared database
function View.static:Configure()
	local views = DB.GetRecord("views", "default")
	for _, name in pairs(views) do
		local v = DB.GetRecord("views", name)
		local c = Puffin.Classes[v.class]
		if c then
			-- Puffin.Debug("Configure View:", name)
			local view = c:new(name, v)
			if view then View.activeViews[name] = view end -- instantiate each default view and save a reference
		end
	end
end

-- Class method to reconfigure views based on configuration info in the shared database
function View.static:UpdateConfiguration()
	local views = View.activeViews
	if views then
		for name, view in pairs(views) do
			local dt = DB.GetRecord("views", name) -- get updated configuration info
			if dt then View.Reconfigure(view, dt) end -- verify description table and then apply changes to existing views
		end
	end
end

Puffin.RegisterCallback("Puffin_Configure", View.Configure)
Puffin.RegisterCallback("Puffin_Reconfigure", View.UpdateConfiguration)
